package at.asit.project.timing;

import android.content.ContextWrapper;
import android.util.Log;

import androidx.annotation.NonNull;

import org.yaml.snakeyaml.Yaml;

import java.io.InputStream;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.LinkedHashMap;

public class Config {
  static final int methods_file_resource = R.raw.methods_04_02;
  static final int constructors_file_resource = R.raw.constructors_04_02;
  // used for return value exploration
  static final int all_methods_file_resource = R.raw.methods_04_02;
  static final int all_constructors_file_resource = R.raw.constructors_04_02;
  private static final String TAG = Config.class.getName();
  public static int NR_OF_INIT_LOOPS;
  public static boolean USE_STRING_STREAM;
  public static boolean FILTER_SAME_VALUES;
  public static boolean FILTER_DOUBLE_METHODS;
  public static boolean SINGLE_OUTPUT_FILE;
  static boolean USE_ONLY_METHODS_FROM_FILE;
  static int EXPLORATION_DEPTH_FOR_CLASS_RETURN_VALUES;
  static ArrayList<String> METHOD_BLACKLIST;
  static ArrayList<String> CONTAINER_CLASSES;
  private static String[] METHOD_PREFIXES;

  public static boolean isInBlacklist(String methodName, String className) {
    return METHOD_BLACKLIST.contains(methodName) || METHOD_BLACKLIST.contains(className + "." + methodName);
  }

  public static void loadConfig(ContextWrapper contextWrapper) {
    InputStream input = contextWrapper.getResources().openRawResource(R.raw.config);
    Yaml yaml = new Yaml();
    LinkedHashMap<String, Object> config = (LinkedHashMap<String, Object>) ((LinkedHashMap<String, Object>) yaml.load(input)).get("config");
    USE_ONLY_METHODS_FROM_FILE = (boolean) config.get("USE_ONLY_METHODS_FROM_FILE");
    EXPLORATION_DEPTH_FOR_CLASS_RETURN_VALUES = (int) config.get("EXPLORATION_DEPTH_FOR_CLASS_RETURN_VALUES");
    ArrayList<String> methodPrefixesAL = ((ArrayList<String>) config.get("METHOD_PREFIXES"));
    METHOD_PREFIXES = new String[methodPrefixesAL.size()];
    for (int i = 0; i < methodPrefixesAL.size(); ++i) {
      METHOD_PREFIXES[i] = methodPrefixesAL.get(i);
    }
    METHOD_BLACKLIST = (ArrayList<String>) config.get("METHOD_BLACKLIST");
    NR_OF_INIT_LOOPS = (int) config.get("NR_OF_INIT_LOOPS");
    USE_STRING_STREAM = (boolean) config.get("USE_STRING_STREAM");
    FILTER_SAME_VALUES = (boolean) config.get("FILTER_SAME_VALUES");
    FILTER_DOUBLE_METHODS = (boolean) config.get("FILTER_DOUBLE_METHODS");
    SINGLE_OUTPUT_FILE = (boolean) config.get("SINGLE_OUTPUT_FILE");
    loadContainerClasses(contextWrapper);
  }

  private static void loadContainerClasses(ContextWrapper contextWrapper) {
    InputStream input = contextWrapper.getResources().openRawResource(R.raw.container_classes);
    Yaml yaml = new Yaml();
    CONTAINER_CLASSES = ((LinkedHashMap<String, ArrayList<String>>) yaml.load(input)).get("container_classes");
    Log.v(TAG, "CONTAINER_CLASSES = " + CONTAINER_CLASSES);
  }

  public static boolean isRelevantMethod(String methodName) {
    for (String methodPrefix : METHOD_PREFIXES) {
      if (methodName.startsWith(methodPrefix)) {
        return true;
      }
    }
    return false;
  }

  public static boolean isPossibleCallback(@NonNull String className, Method[] methods) {
    if (className.contains("Listener")) {
      return true;
    }
    for (Method method : methods) {
      if (method.getName().startsWith("on")) {
        return true;
      }
    }
    return false;
  }

}
