package at.asit.project.timing;

import androidx.annotation.NonNull;

import java.util.Arrays;

class MethodReturn {
    private Object returnValue;
    private long[] timings;
    private Object[] originalObjects;
    private Object[] modifiedObjects;

    MethodReturn(Object returnValue, long[] timings, Object[] originalObjects, Object[] modifiedObjects) {
      this.returnValue = returnValue;
      this.timings = timings;
      this.originalObjects = originalObjects;
      this.modifiedObjects = modifiedObjects;
    }

    Object getActualReturnValue() {
      return returnValue;
    }

    public Object[] getOriginalObjects() {
      return originalObjects;
    }

    public Object[] getModifiedObjects() {
      return modifiedObjects;
    }

    public long[] getTimings() {
      return timings;
    }

    public AveragedReturnTimes getAverageTiming() {
      AveragedReturnTimes averagedReturnTimes = new AveragedReturnTimes();
      averagedReturnTimes.originalRun = Arrays.stream(Arrays.copyOfRange(timings, 0, timings.length / 2)).average().getAsDouble();
      averagedReturnTimes.adaptedRun = Arrays.stream(Arrays.copyOfRange(timings, timings.length / 2, timings.length)).average().getAsDouble();
      averagedReturnTimes.percentage = ((averagedReturnTimes.originalRun / averagedReturnTimes.adaptedRun) - 1.0);
      return averagedReturnTimes;
    }

    @NonNull
    @Override
    public String toString() {
      AveragedReturnTimes averagedReturnTimes = getAverageTiming();
      return "{\nDifference:" + averagedReturnTimes.percentage + "\n  Original Run: " + averagedReturnTimes.originalRun + " ||| Parameter: " + Arrays.toString(originalObjects) + " ||||||| Raw Timings: " + Arrays
              .toString(Arrays.copyOfRange(timings, 0, timings.length / 2)) + "\n  Adapted Run: " + averagedReturnTimes.adaptedRun + " ||| Parameter: " + Arrays.toString(modifiedObjects) + " ||||||| Raw Timings: " + Arrays
              .toString(Arrays.copyOfRange(timings, timings.length / 2, timings.length));
    }

    class AveragedReturnTimes {
      double originalRun;
      double adaptedRun;
      double percentage;

      @NonNull
      @Override
      public String toString() {
        return "AveragedReturnTimes{" + "originalRun=" + originalRun + ", adaptedRun=" + adaptedRun + ", percentage=" + percentage + '}';
      }
    }

  }