package at.asit.project.timing.types;

import android.content.Context;
import androidx.annotation.NonNull;
import android.util.Log;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.nio.ByteBuffer;
import java.nio.channels.AsynchronousFileChannel;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;

import at.asit.project.timing.Config;

public class APIMethod {
  private static final String TAG = APIMethod.class.getName();
  public static boolean KEEP_MEASUREMENTS = false;
  private static Path SINGLE_FILE;
  private static AsynchronousFileChannel SINGLE_FILE_STREAM;
  private static long currentPosition;
  private static Path SINGLE_FILE_MATCH;
  private static AsynchronousFileChannel SINGLE_FILE_STREAM_MATCH;
  private static long currentPositionMatch;
  private static int FILES_CREATED = 0;
  private final Method method;
  private final Object[] parameterObjects;
  private final String className;
  private final String hierarchy;
  public long firstTimestamp;
  public String previouslyWrittenToFile = "";
  private Object classObject;
  private Object parentClassObject;
  private Method objectCreatorMethod;
  private Object[] objectCreatorMethodParameters;

  private StringBuilder measurementsStringStream = new StringBuilder();
  private int nrOfAppendsToStringStream = 0;
  private boolean addedMethodHierarchy = false;
  private Path measurementsFile;
  private AsynchronousFileChannel measurementsFileStream;
  private boolean recreateObject;

  APIMethod(Context context, Object classObject, Method method, Object[] parameterObjects, String className, String hierarchy) {
    File path = null;
    if (SINGLE_FILE != null) {
      path = context.getExternalFilesDir(null);
    }
    this.method = method;
    this.parameterObjects = parameterObjects;
    this.className = className;
    this.hierarchy = hierarchy;
    this.classObject = classObject;
    createSingleFile(path);
  }

  public static Path getSingleFile(File path) {
    if (SINGLE_FILE == null) {
      SINGLE_FILE = Paths.get(path.getAbsolutePath() + "/single_results_NO_match.txt");
    }
    return SINGLE_FILE;
  }

  public static AsynchronousFileChannel getSingleFileStream() throws IOException {
    if (SINGLE_FILE_STREAM == null) {
      SINGLE_FILE_STREAM = AsynchronousFileChannel.open(SINGLE_FILE, StandardOpenOption.WRITE, StandardOpenOption.CREATE);
    }
    return SINGLE_FILE_STREAM;
  }

  public static Path getSingleFileMatch(File path) {
    if (SINGLE_FILE_MATCH == null) {
      SINGLE_FILE_MATCH = Paths.get(path.getAbsolutePath() + "/single_results_match.txt");
    }
    return SINGLE_FILE_MATCH;
  }

  public static AsynchronousFileChannel getSingleFileStreamMatch() throws IOException {
    if (SINGLE_FILE_STREAM_MATCH == null) {
      SINGLE_FILE_STREAM_MATCH = AsynchronousFileChannel.open(SINGLE_FILE_MATCH, StandardOpenOption.WRITE, StandardOpenOption.CREATE);
    }
    return SINGLE_FILE_STREAM_MATCH;
  }

  public static void writeByteToSingleFileStream(byte[] bytes) throws IOException {
    getSingleFileStream().write(ByteBuffer.wrap(bytes), currentPosition);
    currentPosition += bytes.length;
  }

  private void createSingleFile(File path) {
    measurementsFile = getSingleFile(path);
    getSingleFileMatch(path);
    try {
      measurementsFileStream = getSingleFileStream();
      getSingleFileStreamMatch();
    } catch (IOException e) {
      e.printStackTrace();
      throw new RuntimeException("Output file could not be created");
    }
  }

  public boolean hasParentCreators() {
    return parentClassObject != null || objectCreatorMethod != null || objectCreatorMethodParameters != null;
  }

  public void setParentCreators(Object parentClassObject, Method objectCreatorMethod, Object[] objectCreatorMethodParameters) {
    this.parentClassObject = parentClassObject;
    this.objectCreatorMethod = objectCreatorMethod;
    this.objectCreatorMethodParameters = objectCreatorMethodParameters;
  }

  public Object getClassObject() {
    if (objectCreatorMethod != null) {
      try {
        Log.v(TAG, "Recreating class object for " + className);
        classObject = objectCreatorMethod.invoke(parentClassObject, objectCreatorMethodParameters);
      } catch (IllegalAccessException | InvocationTargetException e) {
        e.printStackTrace();
        throw new RuntimeException("Could not recreate class object");
      }
    }
    return classObject;
  }

  public void addTargetToStringStream(String target) {
    writeMeasurement("\n");
  }

  public void addMeasurement(String measurement, boolean match) {
    try {
      if (match) {
        measurementsFileStream = getSingleFileStreamMatch();
      } else {
        measurementsFileStream = getSingleFileStream();
      }
    } catch (IOException e) {
      throw new RuntimeException();
    }
    writeMeasurement(":::" + measurement);
  }

  private void writeMeasurement(String data) {
    ++nrOfAppendsToStringStream;
    if (!KEEP_MEASUREMENTS) {
      return;
    }
    try {
      if (!Config.USE_STRING_STREAM && measurementsFile == null) {
        createFiles();
      }

      if (!addedMethodHierarchy) {
        if (Config.USE_STRING_STREAM) {
          measurementsStringStream.append(hierarchy);
        } else {
          byte[] bytes = hierarchy.getBytes();
          measurementsFileStream.write(ByteBuffer.wrap(bytes), currentPosition);
          currentPosition += bytes.length;
        }
        addedMethodHierarchy = true;

        data = data.replace("\n", "-NEWLINE-");
        data += "\n";
        if (Config.USE_STRING_STREAM) {
          measurementsStringStream.append(data);
        } else {
          byte[] bytes = data.getBytes();
          measurementsFileStream.write(ByteBuffer.wrap(bytes), currentPosition);
          currentPosition += bytes.length;
        }
      }
      //TODO filter changing methods

    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  @NonNull
  private String getSignature() {
    StringBuilder signature = new StringBuilder();
    signature.append(className);
    signature.append("_");
    signature.append(method.getName());
    signature.append("(");
    for (int counter = 0; counter < parameterObjects.length; ++counter) {
      Object parameterObject = parameterObjects[counter];
      if (counter > 0) {
        signature.append(",");
      }
      signature.append(parameterObject.getClass().getName());
    }
    signature.append(")");
    return signature.toString();
  }

  public void writeToFile() {
    try {
      if (measurementsFile == null) {
        createFiles();
      }
      byte[] bytes = measurementsStringStream.toString().getBytes();
      measurementsFileStream.write(ByteBuffer.wrap(bytes), currentPosition);
      currentPosition += bytes.length;
      clearStringStream();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private void createFiles() throws FileNotFoundException {
    ++FILES_CREATED;
    throw new RuntimeException(); // Fix
  }

  public Method getMethod() {
    return method;
  }

  private Object[] recreateParameterObjects() {
    return null;
  }

  public Object[] getParameterObjects() {
    return recreateObject ? recreateParameterObjects() : parameterObjects;
  }

  private void clearStringStream() {
    measurementsStringStream = new StringBuilder();
    nrOfAppendsToStringStream = 0;
  }

  public boolean stringStreamChangedMoreThanOnce() {
    return nrOfAppendsToStringStream > 1;
  }

  @Override
  public String toString() {
    return "method: " + method.getName() + " of class " + className;
  }
}
